
#include "sdbInt.h"

static SdbVariable     *columnOfInterest;

static int              cmpCodeField _ANSI_ARGS_((CONST void *first, 
						  CONST void *last));
static int              cmpNumberField _ANSI_ARGS_((CONST void *first, 
						    CONST void *last));
static void             SdbResetDB _ANSI_ARGS_((SdbFile *dbPtr, int nrecs));
static int              codeSearch _ANSI_ARGS_((SdbVariable *v, int low, 
			       int high, char *key, int *list));
static int              numberSearch _ANSI_ARGS_((SdbVariable *v, int low, 
			       int high, double value, int *list));

static int
cmpCodeField(first, last)
   CONST  void              *first,
                            *last;
{
    return (strcmp(columnOfInterest->u.ivalue[*(int*)first],
                    columnOfInterest->u.ivalue[*(int*)last]));
}

static int
cmpNumberField(first, last)
   CONST  void              *first,
                            *last;
{
    return (columnOfInterest->u.dvalue[*(int*)first] -
            columnOfInterest->u.dvalue[*(int*)last]);
}

void
SdbCloseDB(dbPtr)
     SdbFile                *dbPtr;
{

    SdbVariable            *varPtr,
                           *next;
    SdbStack               *top;

    if (!dbPtr)
        return;

    for (varPtr = dbPtr->varPtr; varPtr; varPtr = next) {
        next = varPtr->next;
        SdbDelete(dbPtr->interp, dbPtr, varPtr->id, 1);
    }

    Tcl_ResetResult(dbPtr->interp);

    if (dbPtr->list)
        ckfree(dbPtr->list);
    if (dbPtr->base)
        ckfree(dbPtr->base);
    if (dbPtr->reverse)
        ckfree(dbPtr->reverse);
    if (dbPtr->dbname)
        ckfree(dbPtr->dbname);
    if (dbPtr->filename)
        ckfree(dbPtr->filename);
    if (dbPtr->outputPtrPtr)
        ckfree(dbPtr->outputPtrPtr);
    if (dbPtr->select)
        ckfree(dbPtr->select);
    if (dbPtr->traceId)
        ckfree(dbPtr->traceId);

    top = dbPtr->stack;
    while (top) {
        SdbStack               *nextStack;

        nextStack = top->next;
        ckfree(top->select);
        ckfree(top);
        top = nextStack;
    }

    if (dbPtr->yieldList) {
        Tcl_HashEntry          *entryPtr;
        Tcl_HashSearch          search;

        /* visit each yield set and delete individually */
        for (entryPtr = Tcl_FirstHashEntry(dbPtr->yieldList, &search);
                entryPtr != NULL;
                entryPtr = Tcl_NextHashEntry(&search)) {
            SdbDeleteYieldSet((SdbYieldSet *) Tcl_GetHashValue(entryPtr));
        }
        Tcl_DeleteHashTable(dbPtr->yieldList);
        ckfree(dbPtr->yieldList);
    }
    if (dbPtr->joinList) {
        Tcl_DeleteHashTable(dbPtr->joinList);
        ckfree(dbPtr->joinList);
    }
    if (dbPtr->randomList) {
        Tcl_HashEntry          *entryPtr;
        Tcl_HashSearch          search;
        SdbRandom              *randomPtr;

        /* visit each random stream and delete individually */
        for (entryPtr = Tcl_FirstHashEntry(dbPtr->randomList, &search);
                entryPtr != NULL;
                entryPtr = Tcl_NextHashEntry(&search)) {
            randomPtr = (SdbRandom *) Tcl_GetHashValue(entryPtr);
            ckfree(randomPtr->state);
            ckfree(randomPtr);
        }
        Tcl_DeleteHashTable(dbPtr->randomList);
        ckfree(dbPtr->randomList);
    }
    if (dbPtr->htbl) {
        Tcl_DeleteHashTable(dbPtr->htbl);
        ckfree(dbPtr->htbl);
    }
    ckfree(dbPtr);
}


/* 
 * The following routine resets the dbPtr structure in preparation to load 
 * in another file.  
 */
static void
SdbResetDB(dbPtr, nrecs)
     SdbFile                *dbPtr;
     int                     nrecs;
{
    int                     i;
    SdbStack               *top,
                           *next;

    dbPtr->position = 0;
    dbPtr->numRec = nrecs;
    dbPtr->numSelect = nrecs;

    /* reallocate the current selection mask */
    if (dbPtr->select)
        ckfree(dbPtr->select);
    dbPtr->select = (_bit_type *) ckalloc(_bit_size(nrecs) * sizeof(_bit_type));

    bit_setall(dbPtr->select, nrecs);

    /* clean out the selection stack */
    for (top = dbPtr->stack; top; top = next) {
        next = top->next;
        ckfree(top->select);
        ckfree(top);
    }

    /* rebuild the record pointer lists */
    if (dbPtr->list)
        ckfree(dbPtr->list);
    dbPtr->list = (int *) ckalloc(nrecs * sizeof(int));

    if (dbPtr->reverse)
        ckfree(dbPtr->reverse);
    dbPtr->reverse = (int *) ckalloc(nrecs * sizeof(int));

    if (dbPtr->base)
        ckfree(dbPtr->base);
    dbPtr->base = (int *) ckalloc(nrecs * sizeof(int));

    for (i = 0; i < dbPtr->numRec; i++) {
        dbPtr->list[i] = i;
        dbPtr->base[i] = i;
        dbPtr->reverse[i] = i;
    }

    /* increment the cache counter in order to make sure that the cache
     * gets flushed out */
    dbPtr->stamp++;
    return;

}

SdbFile                *
SdbCreateDB(interp, dbname, nrecs)
     Tcl_Interp             *interp;
     char                   *dbname;
     int                     nrecs;
{

    /* allocate and initialize a database structure */
    SdbFile                *dbPtr;

    dbPtr = (SdbFile *) ckalloc(sizeof(SdbFile));

    if (!dbPtr)
        return (NULL);

    dbPtr->interp = interp;
    dbPtr->dbname = (char *) ckalloc(strlen(dbname) + 1);
    strcpy(dbPtr->dbname, dbname);
    dbPtr->htbl = (Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));

    Tcl_InitHashTable(dbPtr->htbl, TCL_STRING_KEYS);
    dbPtr->filename = NULL;
    dbPtr->varPtr = NULL;
    dbPtr->outputPtrPtr = NULL;
    dbPtr->numOutput = 0;
    dbPtr->errors=0;
    dbPtr->indexPtr = NULL;
    dbPtr->stamp = 2;
    dbPtr->select = NULL;
    dbPtr->stack = NULL;
    dbPtr->yieldList = (Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));

    Tcl_InitHashTable(dbPtr->yieldList, TCL_STRING_KEYS);
    dbPtr->joinList = (Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));

    Tcl_InitHashTable(dbPtr->joinList, TCL_STRING_KEYS);
    dbPtr->randomList = (Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));

    Tcl_InitHashTable(dbPtr->randomList, TCL_STRING_KEYS);
    dbPtr->list = NULL;
    dbPtr->reverse = NULL;
    dbPtr->base = NULL;
    gettimeofday(&dbPtr->timePtr, NULL);
    SdbResetDB(dbPtr, nrecs);
    return (dbPtr);
}

int
SdbReloadFile(interp, dbPtr, string)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     char                   *string;
{

    SdbVariable            *varPtr;

    FILE                   *infile;

    /* make sure that the file exists and can be opened */
    infile = fopen(string, "r");
    if (!infile) {
        Tcl_AppendResult(interp, "could not open file: ", string,
                (char *) NULL);
        return TCL_ERROR;
    }
    fclose(infile);

    /* get rid of the old fields */
    for (varPtr = dbPtr->varPtr; varPtr; varPtr = varPtr->next)
        if (varPtr->flag & SDB_ALLOCATED)
            SdbDelete(interp, dbPtr, varPtr->id, 1);

    /* now do the work of reloading the file */
    if (SdbOpenFile(interp, string, "reload", dbPtr))
        return TCL_OK;
    else
        return TCL_ERROR;

}

SdbFile                *
SdbOpenFile(interp, string, dbname, dbPtr)
     Tcl_Interp             *interp;
     char                   *string;
     char                   *dbname;
     SdbFile                *dbPtr;
{
    SdbTextFile            *file;
    int                     i,
                            n;
    FILE                   *infile;
    int                     estimate;
    SdbVariable           **varPtr = NULL;

    /* First, obtain an estimate of the number of 
     * records in the input file.  This will be used
     * to allocate space for the stand data  */

    infile = fopen(string, "r");
    estimate = 0;
    if (infile) {
        char                    record[1000];

        setvbuf(infile, NULL, _IOFBF, 8192);
        while (fgets(record, 1000, infile))
            if (strncmp("DATA", record, 4) == 0)
                estimate++;
        fclose(infile);
    } else {
        Tcl_AppendResult(interp, "could not open file: ", string,
                (char *) NULL);
        return NULL;
    }

    /* Open up the inventory file using the textfile routines.
     * If this fails, return immediately */
    if (!(file = SdbTextOpen(interp, string, 8192))) {
        Tcl_AppendResult(interp, "file open failure: ", string,
                (char *) NULL);
        return NULL;
    }
    /* if required, allocate and partially initialize the database
     * structure */
    if (!dbPtr)
        dbPtr = SdbCreateDB(interp, dbname, estimate);
    if (!dbPtr) {
        Tcl_AppendResult(interp, "could not allocate file structure", NULL);
        SdbTextClose(file);
        return NULL;
    }
    /* now initialize all of the rest of the fields */
    SdbResetDB(dbPtr, estimate);

    /* go through the variables in the input file, 
     * allocate the array,
     * and add them into the active variable structure */
    varPtr = (SdbVariable **) ckalloc(file->n * sizeof(SdbVariable *));

    for (i = 0; i < file->n; i++) {
        if (SdbFound(dbPtr->varPtr, file->field[i]))
            SdbDelete(interp, dbPtr, file->field[i],1);
        varPtr[i] = SdbInsert(dbPtr, file->field[i],
                file->present[i] | SDB_OUTPUT | SDB_ALLOCATED, estimate);
    }
    /* tidy up */
    SdbReOptimize(dbPtr);

    /* loop through reading in records.
     * text_get will parse the input records and put everything
     * into an array named file->list.  All we have to do is
     * move the data from the list into the stand stucture, and
     * convert to the correct code format. */

    n = 0;
    while (SdbTextGet(file)) {

        int                     new;
        double                  number;
        char                    tempString[10];

        /* Now load up all of the variables from the file */
        for (i = 0; i < file->n; i++) {
            Tcl_HashEntry          *entryPtr;
            SdbVariable            *v = varPtr[i];

            if (v->flag & SDB_CODE) {
                entryPtr = Tcl_CreateHashEntry(dbPtr->htbl,
                        file->list[i], &new);
                v->u.ivalue[n] = entryPtr->key.string;
                Tcl_SetHashValue(entryPtr, n);
            } else if (strcmp(file->list[i], ".") == 0)
                v->u.dvalue[n] = 0.0;
            else if (!sscanf(file->list[i], "%lf", &number)) {
                sprintf(tempString, "%d", n);
                Tcl_AppendResult(interp, "Invalid number <", file->list[i],
                        "> in field ", v->id, ", record ", tempString,
                        (char *) NULL);
                number = 0.0;
            } else
                v->u.dvalue[n] = number;
        }

        n++;
    }
    ckfree(varPtr);

    SdbTextClose(file);
    return dbPtr;

}

static int
codeSearch(v, low, high, key, list)
     SdbVariable            *v;
     int                     low;
     int                     high;
     char                   *key;
     int                    *list;
{
    int                     result;
    int                     mid;

    while (low <= high) {
        mid = (low + high) / 2;

        result = strcmp(key, v->u.ivalue[list[mid]]);

        if (result < 0)
            high = mid - 1;
        else if (result > 0)
            low = mid + 1;
        else
            return (list[mid]);
    }

    return (-1);
}
static int
numberSearch(v, low, high, key, list)
     SdbVariable            *v;
     int                     low;
     int                     high;
     double                  key;
     int                    *list;
{
    double                  result;
    int                     mid;

    while (low <= high) {
        mid = (low + high) / 2;

        result = key - v->u.dvalue[list[mid]];

        if (result < 0)
            high = mid - 1;
        else if (result > 0)
            low = mid + 1;
        else
            return (list[mid]);
    }

    return (-1);
}

int
SdbSeek(interp, dbPtr, key)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     char                   *key;
{

    SdbVariable            *v = dbPtr->indexPtr;

    if (!v) {
        Tcl_SetResult(interp, "database not indexed", TCL_STATIC);
        return -1;
    }
    if (v->flag & SDB_CODE)
        return (dbPtr->reverse[codeSearch(v, 0, dbPtr->numRec, key, dbPtr->list)]);
    else {
        double                  value;

        if (sscanf(key, "%lf", &value))
            return (dbPtr->reverse[numberSearch(v, 0, dbPtr->numRec, value, dbPtr->list)]);
        else
            return -1;
    }

}

int
SdbIndex(interp, dbPtr, field)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     char                   *field;
{

    SdbVariable            *varPtr;

    if (!(varPtr = SdbFound(dbPtr->varPtr, field))) {
        Tcl_AppendResult(interp, "field not found \"",
                field, "\"", NULL);
        return TCL_ERROR;
    }
    if (varPtr->flag & (SDB_EXPRESSION | SDB_LOOKUP)) {
        Tcl_AppendResult(interp, 
		"cannot index on expression or lookup field \"",
                field, "\"", NULL);
        return TCL_ERROR;
    }
    if (dbPtr->indexPtr)
        (dbPtr->indexPtr)->flag ^= SDB_INDEXED;
    varPtr->flag |= SDB_INDEXED;
    dbPtr->indexPtr = varPtr;

    columnOfInterest = dbPtr->indexPtr;
    qsort(dbPtr->list, (unsigned) dbPtr->numRec, sizeof(int),
	  ((dbPtr->indexPtr)->flag & SDB_CODE) ? cmpCodeField : 
	  cmpNumberField);

    Tcl_AppendResult(interp, field, NULL);
    return TCL_OK;

}

int
SdbSummary(interp, dbPtr, argc, argv)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     int                     argc;
     char                  **argv;
{

    int                    *keys,
                            i,
                            j,
                            k,
                            l,
                            rec,
                            new,
                            cats = 0,
                            nums = 0;
    Tcl_HashTable           hashTable;
    Tcl_HashEntry          *entryPtr;
    double                 *values;
    Tcl_HashSearch          search;
    typedef struct field_type {
        SdbVariable            *var;
        int                     type;
    } field_type;
    field_type             *list;

    /* first, figure out how many categories and numerics */
    list = (field_type *) ckalloc(argc * sizeof(field_type));

    for (i = 0; i < argc; i++) {
        if ((list[i].var = SdbFound(dbPtr->varPtr, argv[i])) == NULL) {
            Tcl_AppendResult(interp, "variable \"", argv[i], "\" not found", NULL);
            ckfree(list);
            return TCL_ERROR;
        }
        if ((list[i].var)->flag & SDB_CODE) {
            cats++;
            list[i].type = 0;
        } else {
            nums++;
            list[i].type = 1;
        }
    }

    if (cats == 0) {
        Tcl_AppendResult(interp, "no categories selected", NULL);
        ckfree(list);
        return TCL_ERROR;
    }
    if (nums == 0) {
        Tcl_AppendResult(interp, "no numerics selected", NULL);
        ckfree(list);
        return TCL_ERROR;
    }
    /* create the hash table and an array to hold the keys */
    if (cats == 1)
        cats = 2;               /* tcl does not like 1 word hash arrays */
    Tcl_InitHashTable(&hashTable, cats);
    keys = (int *) ckalloc(cats * sizeof(int));

    keys[1] = 0;

    /* loop through selected records */
    for (j = 0; j < dbPtr->numRec; j++) {
        rec = dbPtr->base[j];
        if (bit_test(dbPtr->select, rec)) {
            dbPtr->stamp++;

            /* build the hash key and attempt a retrieval */
            k = 0;
            for (i = 0; i < argc; i++) {
                if (!list[i].type) {
                    keys[k] = (int) SdbGetCode(dbPtr, list[i].var, rec);
                    k++;
                }
            }
            entryPtr = Tcl_CreateHashEntry(&hashTable, (char *) keys, &new);

            /* if no previous entry then add a new one */
            if (new) {
                values = (double *) ckalloc(nums * sizeof(double));

                for (l = 0; l < nums; l++)
                    values[l] = 0.0;
                Tcl_SetHashValue(entryPtr, values);
            } else
                values = (double *) Tcl_GetHashValue(entryPtr);

            /* now increment the values in the hash table */
            l = 0;
            for (i = 0; i < argc; i++) {
                if (list[i].type) {
                    values[l] += SdbGetNumber(dbPtr, list[i].var, rec);
                    l++;
                }
            }
        }
    }
    ckfree(keys);

    /* Retrieve the values from the hash table */
    for (entryPtr = Tcl_FirstHashEntry(&hashTable, &search);
            entryPtr != NULL;
            entryPtr = Tcl_NextHashEntry(&search)) {
        if (interp->result && *interp->result)
	    Tcl_AppendResult(interp, " ", NULL);	  
        Tcl_AppendResult(interp, "{", NULL);
        keys = (int *) Tcl_GetHashKey(&hashTable, entryPtr);
        values = (double*) Tcl_GetHashValue(entryPtr);
        k = 0;
        l = 0;
        for (i = 0; i < argc; i++) {
            if (!list[i].type) {
                Tcl_AppendElement(interp, (char *) keys[k]);
                k++;
            } else {
                char                    string[64];

                sprintf(string, (list[i].var)->format, values[l]);
                Tcl_AppendElement(interp, string);
                l++;
            }
        }
        Tcl_AppendResult(interp, "}", NULL);
        Tcl_DeleteHashEntry(entryPtr);
        ckfree(values);
    }

    ckfree(list);
    Tcl_DeleteHashTable(&hashTable);
    return TCL_OK;
}
